from datetime import datetime
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Tuple,
    TypeVar,
)

import numpy as np
import pandas as pd
import pyarrow as pa
from pandas.core.frame import DataFrame

from triad.utils.assertion import assert_or_throw
from triad.utils.pyarrow import (
    TRIAD_DEFAULT_TIMESTAMP_UNIT,
    cast_pa_table,
    pa_table_to_pandas,
    to_pa_datatype,
    to_pandas_dtype,
)

T = TypeVar("T", bound=Any)
ColT = TypeVar("ColT", bound=Any)

_DEFAULT_DATETIME = datetime(2000, 1, 1)
_ANTI_INDICATOR = "__anti_indicator__"
_CROSS_INDICATOR = "__corss_indicator__"


class PandasLikeUtils(Generic[T, ColT]):
    """A collection of utils for general pandas like dataframes"""

    def empty(self, df: T) -> bool:
        """Check if the dataframe is empty

        :param df: pandas like dataframe
        :return: if it is empty
        """
        return len(df.index) == 0

    def as_arrow(self, df: T, schema: Optional[pa.Schema] = None) -> pa.Table:
        """Convert pandas like dataframe to pyarrow table

        :param df: pandas like dataframe
        :param schema: if specified, it will be used to construct pyarrow table,
          defaults to None

        :return: pyarrow table
        """
        return pa.Table.from_pandas(df, schema=schema, preserve_index=False, safe=False)

    def as_array_iterable(
        self,
        df: T,
        schema: Optional[pa.Schema] = None,
        columns: Optional[List[str]] = None,
        type_safe: bool = False,
    ) -> Iterable[List[Any]]:
        """Convert pandas like dataframe to iterable of rows in the format of list.

        :param df: pandas like dataframe
        :param schema: schema of the input. With None, it will infer the schema,
          it can't infer wrong schema for nested types, so try to be explicit
        :param columns: columns to output, None for all columns
        :param type_safe: whether to enforce the types in schema, if False, it will
            return the original values from the dataframe
        :return: iterable of rows, each row is a list
        """
        if self.empty(df):
            return
        if schema is None:
            schema = self.to_schema(df)
        if columns is not None:
            df = df[columns]
            schema = pa.schema([schema.field(n) for n in columns])
        if not type_safe:
            for arr in df.itertuples(index=False, name=None):
                yield list(arr)
        else:
            p = self.as_arrow(df, schema)
            d = p.to_pydict()
            cols = [d[n] for n in schema.names]
            for arr in zip(*cols):
                yield list(arr)

    def to_schema(self, df: T) -> pa.Schema:
        """Extract pandas dataframe schema as pyarrow schema. This is a replacement
        of pyarrow.Schema.from_pandas, and it can correctly handle string type and
        empty dataframes

        :param df: pandas dataframe
        :raises ValueError: if pandas dataframe does not have named schema
        :return: pyarrow.Schema

        :Notice:
        The dataframe must be either empty, or with type pd.RangeIndex, pd.Int64Index
        or pd.UInt64Index and without a name, otherwise, `ValueError` will raise.
        """
        self.ensure_compatible(df)
        assert_or_throw(
            df.columns.dtype == "object",
            ValueError("Pandas dataframe must have named schema"),
        )

        def get_fields() -> Iterable[pa.Field]:
            if isinstance(df, pd.DataFrame) and len(df.index) > 0:
                yield from pa.Schema.from_pandas(df, preserve_index=False)
            else:
                for i in range(df.shape[1]):
                    tp = df.dtypes.iloc[i]
                    if tp == np.dtype("object") or pd.api.types.is_string_dtype(tp):
                        t = pa.string()
                    elif isinstance(tp, pd.DatetimeTZDtype):
                        t = pa.timestamp(tp.unit, str(tp.tz))
                    else:
                        t = to_pa_datatype(tp)
                    yield pa.field(df.columns[i], t)

        fields: List[pa.Field] = []
        for field in get_fields():
            if pa.types.is_timestamp(field.type):
                fields.append(
                    pa.field(
                        field.name,
                        pa.timestamp(TRIAD_DEFAULT_TIMESTAMP_UNIT, field.type.tz),
                    )
                )
            elif pa.types.is_large_string(field.type):
                fields.append(pa.field(field.name, pa.string()))
            else:
                fields.append(field)
        return pa.schema(fields)

    def cast_df(
        self,
        df: T,
        schema: pa.Schema,
        use_extension_types: bool = True,
        use_arrow_dtype: bool = False,
        **kwargs: Any,
    ) -> T:
        """Cast pandas like dataframe to comply with ``schema``.

        :param df: pandas like dataframe
        :param schema: pyarrow schema to cast to
        :param use_extension_types: whether to use ``ExtensionDType``, default True
        :param use_arrow_dtype: whether to use ``ArrowDtype``, default False
        :param kwargs: other arguments passed to ``pa.Table.from_pandas``

        :return: converted dataframe
        """
        dtypes = to_pandas_dtype(
            schema,
            use_extension_types=use_extension_types,
            use_arrow_dtype=use_arrow_dtype,
        )
        if len(df) == 0:
            return pd.DataFrame({k: pd.Series(dtype=v) for k, v in dtypes.items()})
        if dtypes == df.dtypes.to_dict():
            return df
        adf = pa.Table.from_pandas(
            df, preserve_index=False, safe=False, **{"nthreads": 1, **kwargs}
        ).replace_schema_metadata()
        adf = cast_pa_table(adf, schema)
        return pa_table_to_pandas(
            adf,
            use_extension_types=use_extension_types,
            use_arrow_dtype=use_arrow_dtype,
        )

    def to_parquet_friendly(
        self, df: T, partition_cols: Optional[List[str]] = None
    ) -> T:
        """Parquet doesn't like pd.ArrowDtype(<nested types>), this function
        converts all nested types to object types

        :param df: the input dataframe
        :param partition_cols: the partition columns, if any, default None
        :return: the converted dataframe
        """
        pcols = partition_cols or []
        changed = False
        new_types: Dict[str, Any] = {}
        for k, v in df.dtypes.items():
            if k in pcols:
                new_types[k] = np.dtype(object)
                changed = True
            elif (
                hasattr(pd, "ArrowDtype")
                and isinstance(v, pd.ArrowDtype)
                and pa.types.is_nested(v.pyarrow_dtype)
            ):
                new_types[k] = np.dtype(object)
                changed = True
            else:
                new_types[k] = v
        if changed:
            df = df.astype(new_types)
        return df

    def safe_groupby_apply(
        self,
        df: T,
        cols: List[str],
        func: Callable[[T], T],
        key_col_name="__safe_groupby_key__",
        **kwargs: Any,
    ) -> T:
        """Safe groupby apply operation on pandas like dataframes.
        In pandas like groupby apply, if any key is null, the whole group is dropped.
        This method makes sure those groups are included.

        :param df: pandas like dataframe
        :param cols: columns to group on, can be empty
        :param func: apply function, df in, df out
        :param key_col_name: temp key as index for groupu.
            default "__safe_groupby_key__"
        :return: output dataframe

        :Notice:
        The dataframe must be either empty, or with type pd.RangeIndex, pd.Int64Index
        or pd.UInt64Index and without a name, otherwise, `ValueError` will raise.
        """

        def _wrapper(df: T) -> T:
            return func(df.reset_index(drop=True))

        self.ensure_compatible(df)
        if len(cols) == 0:
            return func(df)
        return (
            df.groupby(cols, dropna=False, group_keys=False)
            .apply(lambda df: _wrapper(df), **kwargs)
            .reset_index(drop=True)
        )

    def fillna_default(self, col: Any) -> Any:
        """Fill column with default values according to the dtype of the column.

        :param col: series of a pandas like dataframe
        :return: filled series
        """
        dtype = col.dtype
        if pd.api.types.is_datetime64_dtype(dtype):
            return col.fillna(_DEFAULT_DATETIME)
        if pd.api.types.is_string_dtype(dtype):
            return col.fillna("")
        if pd.api.types.is_bool_dtype(dtype):
            return col.fillna(False)
        return col.fillna(0)

    def is_compatile_index(self, df: T) -> bool:
        """Check whether the datafame is compatible with the operations inside
        this utils collection

        :param df: pandas like dataframe
        :return: if it is compatible
        """
        return (
            isinstance(df.index, pd.RangeIndex) or df.index.inferred_type == "integer"
        )

    def ensure_compatible(self, df: T) -> None:
        """Check whether the datafame is compatible with the operations inside
        this utils collection, if not, it will raise ValueError

        :param df: pandas like dataframe
        :raises ValueError: if not compatible
        """
        if df.index.name is not None:
            raise ValueError("pandas like datafame index can't have name")
        if self.is_compatile_index(df):
            return
        if self.empty(df):  # for pandas < 2
            return  # pragma: no cover
        raise ValueError(
            f"pandas like datafame must have default index, but got {type(df.index)}"
        )

    def parse_join_type(self, join_type: str) -> str:
        """Parse join type string to standard join type string

        :param join_type: the join type string
        :return: the standard join type string
        """
        join_type = join_type.replace(" ", "").replace("_", "").lower()
        if join_type in ["inner", "cross"]:
            return join_type
        if join_type in ["inner", "join"]:
            return "inner"
        if join_type in ["semi", "leftsemi"]:
            return "left_semi"
        if join_type in ["anti", "leftanti"]:
            return "left_anti"
        if join_type in ["left", "leftouter"]:
            return "left_outer"
        if join_type in ["right", "rightouter"]:
            return "right_outer"
        if join_type in ["outer", "full", "fullouter"]:
            return "full_outer"
        raise NotImplementedError(join_type)

    def drop_duplicates(self, df: T) -> T:
        """Remove duplicated rows

        :param df: the dataframe
        :return: the dataframe without duplicated rows
        """
        return df.drop_duplicates()

    def union(self, ndf1: T, ndf2: T, unique: bool) -> T:
        """Union two dataframes

        :param ndf1: dataframe 1
        :param ndf2: dataframe 2
        :param unique: whether to remove duplicated rows
        :return: the unioned dataframe
        """
        ndf1, ndf2 = self._preprocess_set_op(ndf1, ndf2)
        ndf = self.concat_dfs(ndf1, ndf2)
        if unique:
            ndf = self.drop_duplicates(ndf)
        return ndf

    def intersect(self, df1: T, df2: T, unique: bool) -> T:
        """Intersect two dataframes

        :param ndf1: dataframe 1
        :param ndf2: dataframe 2
        :param unique: whether to remove duplicated rows
        :return: the intersected dataframe
        """
        ndf1, ndf2 = self._preprocess_set_op(df1, df2)
        ndf = ndf1.merge(self.drop_duplicates(ndf2))
        if unique:
            ndf = self.drop_duplicates(ndf)
        return ndf

    def except_df(
        self,
        df1: T,
        df2: T,
        unique: bool,
        anti_indicator_col: str = _ANTI_INDICATOR,
    ) -> T:
        """Remove df2 from df1

        :param df1: dataframe 1
        :param df2: dataframe 2
        :param unique: whether to remove duplicated rows in the result
        :return: the dataframe with df2 removed
        """
        ndf1, ndf2 = self._preprocess_set_op(df1, df2)
        # TODO: lack of test to make sure original ndf2 is not changed
        ndf2 = self._with_indicator(ndf2, anti_indicator_col)
        ndf = ndf1.merge(ndf2, how="left", on=list(ndf1.columns))
        ndf = ndf[ndf[anti_indicator_col].isnull()].drop([anti_indicator_col], axis=1)
        if unique:
            ndf = self.drop_duplicates(ndf)
        return ndf

    def join(
        self,
        ndf1: T,
        ndf2: T,
        join_type: str,
        on: List[str],
        anti_indicator_col: str = _ANTI_INDICATOR,
        cross_indicator_col: str = _CROSS_INDICATOR,
    ) -> T:
        """Join two dataframes

        :param ndf1: dataframe 1
        :param ndf2: dataframe 2
        :param join_type: join type, can be inner, left_semi, left_anti, left_outer,
            right_outer, full_outer, cross
        :param on: join keys

        :return: the joined dataframe
        """
        join_type = self.parse_join_type(join_type)
        if join_type == "inner":
            ndf1 = ndf1.dropna(subset=on)
            ndf2 = ndf2.dropna(subset=on)
            joined = ndf1.merge(ndf2, how=join_type, on=on)
        elif join_type == "left_semi":
            ndf1 = ndf1.dropna(subset=on)
            ndf2 = self.drop_duplicates(ndf2[on].dropna())
            joined = ndf1.merge(ndf2, how="inner", on=on)
        elif join_type == "left_anti":
            # TODO: lack of test to make sure original ndf2 is not changed
            ndf2 = self.drop_duplicates(ndf2[on].dropna())
            ndf2 = self._with_indicator(ndf2, anti_indicator_col)
            joined = ndf1.merge(ndf2, how="left", on=on)
            joined = joined[joined[anti_indicator_col].isnull()].drop(
                [anti_indicator_col], axis=1
            )
        elif join_type == "left_outer":
            ndf2 = ndf2.dropna(subset=on)
            joined = ndf1.merge(ndf2, how="left", on=on)
        elif join_type == "right_outer":
            ndf1 = ndf1.dropna(subset=on)
            joined = ndf1.merge(ndf2, how="right", on=on)
        elif join_type == "full_outer":
            add: List[str] = []
            for f in on:
                name = f + "_null"
                s1 = ndf1[f].isnull().astype(int)
                ndf1[name] = s1
                s2 = ndf2[f].isnull().astype(int) * 2
                ndf2[name] = s2
                add.append(name)
            joined = ndf1.merge(ndf2, how="outer", on=on + add).drop(add, axis=1)
        elif join_type == "cross":
            assert_or_throw(
                len(on) == 0, ValueError(f"cross join can't have join keys {on}")
            )
            ndf1 = self._with_indicator(ndf1, cross_indicator_col)
            ndf2 = self._with_indicator(ndf2, cross_indicator_col)
            joined = ndf1.merge(ndf2, how="inner", on=[cross_indicator_col]).drop(
                [cross_indicator_col], axis=1
            )
        else:  # pragma: no cover
            raise NotImplementedError(join_type)
        return joined

    def concat_dfs(self, *dfs: T) -> T:  # pragma: no cover
        """Concatenate dataframes

        :param dfs: the dataframes to concatenate
        :return: the concatenated dataframe
        """
        raise NotImplementedError

    def _preprocess_set_op(self, ndf1: T, ndf2: T) -> Tuple[T, T]:
        assert_or_throw(
            len(list(ndf1.columns)) == len(list(ndf2.columns)),
            ValueError("two dataframes have different number of columns"),
        )
        ndf2.columns = ndf1.columns  # this is SQL behavior
        return ndf1, ndf2

    def _with_indicator(self, df: T, name: str) -> T:
        return df.assign(**{name: 1})


class PandasUtils(PandasLikeUtils[pd.DataFrame, pd.Series]):
    """A collection of pandas utils"""

    def concat_dfs(self, *dfs: DataFrame) -> DataFrame:
        return pd.concat(list(dfs))


PD_UTILS = PandasUtils()



from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

from triad.utils.assertion import assert_or_throw
from triad.utils.convert import to_size

T = TypeVar("T")


def make_empty_aware(it: Union[Iterable[T], Iterator[T]]) -> "EmptyAwareIterable[T]":
    """Make an iterable empty aware, or return itself if already empty aware

    :param it: underlying iterable

    :return: EmptyAwareIterable[T]
    """
    return it if isinstance(it, EmptyAwareIterable) else EmptyAwareIterable(it)


def slice_iterable(
    it: Union[Iterable[T], Iterator[T]], slicer: Callable[[int, T, Optional[T]], bool]
) -> "Iterable[_SliceIterable[T]]":
    """Slice the original iterable into slices by slicer

    :param it: underlying iterable
    :param slicer: taking in current number, current value, last value,
        it decides if it's a new slice

    :yield: an iterable of iterables (_SliceIterable[T])
    """
    si = _SliceIterable(it, slicer)
    while si._state < 3:
        yield si
        si.recycle()


def to_kv_iterable(  # noqa: C901
    data: Any, none_as_empty: bool = True
) -> Iterable[Tuple[Any, Any]]:
    """Convert data to iterable of key value pairs

    :param data: input object, it can be a dict or Iterable[Tuple[Any, Any]]
        or Iterable[List[Any]]
    :param none_as_empty: if to treat None as empty iterable

    :raises ValueError: if input is None and `none_as_empty==False`
    :raises ValueError: if input is a set
    :raises TypeError or ValueError: if input data type is not acceptable

    :yield: iterable of key value pair as tuples
    """
    if data is None:
        assert_or_throw(none_as_empty, ValueError("data can't be None"))
    elif isinstance(data, Dict):
        for k, v in data.items():
            yield k, v
    elif isinstance(data, Set):
        raise ValueError(f"{data} is a set, did you mistakenly use `,` instead of `:`?")
    elif isinstance(data, Iterable):
        ei = make_empty_aware(data)
        if not ei.empty:
            first = ei.peek()
            if isinstance(first, tuple):
                for k, v in ei:
                    yield k, v
            elif isinstance(first, List):
                for arr in ei:
                    if len(arr) == 2:
                        yield arr[0], arr[1]
                    else:
                        raise TypeError(f"{arr} is not an acceptable item")
            else:
                raise TypeError(f"{first} is not an acceptable item")
    else:
        raise TypeError(f"{type(data)} is not supported")


class EmptyAwareIterable(Iterable[T]):
    """A wrapper of iterable that can tell if the underlying
    iterable is empty, it can also peek a non-empty iterable.

    :param it: the underlying iterable

    :raises StopIteration: raised by the underlying iterable
    """

    def __init__(self, it: Union[Iterable[T], Iterator[T]]):
        self._last: Optional[T] = None
        if not isinstance(it, Iterator):
            self._iter = iter(it)
        else:
            self._iter = it
        self._state = 0
        self._fill_last()

    @property
    def empty(self) -> bool:
        """Check if the underlying iterable has more items

        :return: whether it is empty
        """
        return self._fill_last() >= 2

    def peek(self) -> T:
        """Return the next of the iterable without moving

        :raises StopIteration: if it's empty
        :return: the `next` item
        """
        if not self.empty:
            return self._last  # type: ignore
        raise StopIteration("Can't peek empty iterable")

    def __iter__(self) -> Any:
        """Wrapper of the underlying __iter__

        :yield: next object
        """
        while self._fill_last() < 2:
            self._state = 0
            yield self._last

    def _fill_last(self) -> int:
        try:
            if self._state == 0:  # last is not filled
                self._last = next(self._iter)
                self._state = 1  # last filed not used
        except StopIteration:
            self._state = 3  # end
        return self._state


class Slicer:
    """A better version of :func:`~triad.iter.slice_iterable`

    :param sizer: the function to get size of an item
    :param row_limit: max row for each slice, defaults to None
    :param size_limit: max byte size for each slice, defaults to None
    :param slicer: taking in current number, current value, last value,
        it decides if it's a new slice

    :raises AssertionError: if `size_limit` is not None but `sizer` is None
    """

    def __init__(
        self,
        sizer: Optional[Callable[[Any], int]] = None,  # func for getsizeof(item)
        row_limit: Optional[int] = None,
        size_limit: Any = None,
        slicer: Optional[Callable[[int, T, Optional[T]], bool]] = None,
    ) -> None:
        self._sizer = sizer
        self._slicer = slicer
        if row_limit is None:
            self._row_limit = 0
        else:
            self._row_limit = row_limit
        if size_limit is None:
            self._size_limit = 0
        else:
            self._size_limit = to_size(str(size_limit))
        assert (
            self._size_limit == 0 or self._sizer is not None
        ), "sizer must be set when size_limit>0"
        self._current_row = 1
        self._current_size = 0

    def slice(  # noqa C901
        self, orig_it: Iterable[T]
    ) -> Iterable[EmptyAwareIterable[T]]:
        """Slice the original iterable into slices by the combined slicing logic

        :param orig_it: ther original iterable

        :yield: an iterable of EmptyAwareIterable
        """
        it = make_empty_aware(orig_it)
        if it.empty:
            pass
        elif self._row_limit <= 0 and self._size_limit <= 0:
            if self._slicer is None:
                yield it
            else:
                for _slice in slice_iterable(it, self._slicer):
                    yield _slice
        elif self._row_limit > 0 and self._size_limit <= 0:
            if self._slicer is None:
                for _slice in slice_iterable(it, self._is_boundary_row_only):
                    yield _slice
            else:
                for _slice in slice_iterable(it, self._is_boundary_row_only_w_slicer):
                    yield _slice
        else:
            self._current_size = self._sizer(it.peek())  # type: ignore
            self._current_row = 1
            if self._row_limit <= 0 and self._size_limit > 0:
                for _slice in slice_iterable(it, self._is_boundary_size_only):
                    yield _slice
            else:
                for _slice in slice_iterable(it, self._is_boundary):
                    yield _slice

    def _is_boundary_row_only(self, no: int, current: Any, last: Any) -> bool:
        return no % self._row_limit == 0

    def _is_boundary_row_only_w_slicer(self, no: int, current: Any, last: Any) -> bool:
        # self._slicer must be invoked even hitting row limit
        is_boundary = self._slicer is not None and self._slicer(no, current, last)
        if self._current_row >= self._row_limit or is_boundary:
            self._current_row = 1
            return True
        self._current_row += 1
        return False

    def _is_boundary_size_only(self, no: int, current: Any, last: Any) -> bool:
        obj_size = self._sizer(current)  # type: ignore
        next_size = self._current_size + obj_size
        # self._slicer must be invoked even hitting size limit
        is_boundary = self._slicer is not None and self._slicer(no, current, last)
        if next_size > self._size_limit or is_boundary:
            self._current_size = obj_size
            return True
        self._current_size = next_size
        return False

    def _is_boundary(self, no: int, current: Any, last: Any) -> bool:
        obj_size = self._sizer(current)  # type: ignore
        next_size = self._current_size + obj_size
        # self._slicer must be invoked even hitting row limit and size limit
        is_boundary = self._slicer is not None and self._slicer(no, current, last)
        if (
            next_size > self._size_limit
            or self._current_row >= self._row_limit  # noqa: W503
            or is_boundary  # noqa: W503
        ):
            self._current_size = obj_size
            self._current_row = 1
            return True
        self._current_size = next_size
        self._current_row += 1
        return False


class _SliceIterable(EmptyAwareIterable[T]):
    def __init__(self, it: Union[Iterable[T], Iterator[T]], slicer: Any):
        self._n = 0
        self._slicer = slicer
        super().__init__(it)

    def recycle(self) -> None:
        if self._state < 2:
            for _ in self:
                pass
        if self._state == 2:
            self._state = 1

    def _fill_last(self) -> int:
        try:
            if self._state == 0:  # last is not filled
                last = self._last
                self._last = next(self._iter)
                is_boundary = self._n > 0 and self._slicer(self._n, self._last, last)
                self._n += 1
                self._state = 2 if is_boundary else 1  # last filed not used
        except StopIteration:
            self._state = 3  # end
        return self._state



import copy
import inspect
import re
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    get_type_hints,
    no_type_check,
)

from ..exceptions import InvalidOperationError
from ..utils.assertion import assert_or_throw
from ..utils.convert import compare_annotations, get_full_type_path
from ..utils.entry_points import load_entry_point
from ..utils.hash import to_uuid
from .dict import IndexedOrderedDict


class FunctionWrapper:
    """Create a function wrapper that can recognize and validate all
    input types.

    :param func: the function to be wrapped
    :param params_re: paramter types regex expression
    :param return_re: return types regex expression

    .. admonition:: Examples

        Here is a simple example to show how to use FunctionWrapper. Assuming
        we want to validate the functions with 2 pandas dataframes as the first
        two input and then arbitray other input, and with 1 pandas dataframe
        as the return

        .. code-block:: python

            import pandas as pd

            @function_wrapper(None)  # all param defintions are here, no entrypoint
            class MyFuncWrapper(FunctionWrapper):
                def __init__(self, func):
                    super().__init__(
                        func,
                        params_re="^dd.*",  # starts with two dataframe parameters
                        return_re="^d$",  # returns a dataframe
                    )

            @MyFuncWrapper.annotated_param(pd.DataFrame, code="d")
            class MyDataFrameParam(AnnotatedParam):
                pass

            def f1(a:pd.DataFrame, b:pd.DataFrame, c) -> pd.DataFrame:
                return a

            def f2(a, b:pd.DataFrame, c):
                return a

            # f1 is valid
            MyFuncWrapper(f1)

            # f2 is invalid because of the first parameter
            # TypeError will be thrown
            MyFuncWrapper(f2)
    """

    _REGISTERED: List[
        Tuple[Type["AnnotatedParam"], Any, str, Callable[[Any], bool]]
    ] = []
    _REGISTERED_CODES: Dict[str, Any] = {}
    _ENTRYPOINT: Optional[str] = None

    def __init__(
        self,
        func: Callable,
        params_re: str = ".*",
        return_re: str = ".*",
    ):
        self._class_method, self._params, self._rt = self._parse_function(
            func, params_re, return_re
        )
        self._func = func

    def __deepcopy__(self, memo: Any) -> Any:
        return copy.copy(self)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self._func(*args, **kwargs)

    def __uuid__(self) -> str:
        return to_uuid(get_full_type_path(self._func), self._params, self._rt)

    @property
    def input_code(self) -> str:
        """The input parameters code expression"""
        return "".join(x.code for x in self._params.values())

    @property
    def output_code(self) -> str:
        """The output code expression"""
        return self._rt.code

    def _parse_function(
        self, func: Callable, params_re: str = ".*", return_re: str = ".*"
    ) -> Tuple[bool, IndexedOrderedDict[str, "AnnotatedParam"], "AnnotatedParam"]:
        sig = inspect.signature(func)
        annotations = get_type_hints(func)
        res: IndexedOrderedDict[str, "AnnotatedParam"] = IndexedOrderedDict()
        class_method = False
        for k, w in sig.parameters.items():
            if k == "self":
                res[k] = SelfParam(w)
                class_method = True
            else:
                anno = annotations.get(k, w.annotation)
                res[k] = self.__class__.parse_annotation(anno, w)
        anno = annotations.get("return", sig.return_annotation)
        rt = self.__class__.parse_annotation(anno, None, none_as_other=False)
        params_str = "".join(x.code for x in res.values())
        assert_or_throw(
            re.match(params_re, params_str) is not None,
            lambda: TypeError(f"Input types not valid {res} for {func}"),
        )
        assert_or_throw(
            re.match(return_re, rt.code) is not None,
            lambda: TypeError(f"Return type not valid {rt} for {func}"),
        )
        return class_method, res, rt

    @no_type_check
    @classmethod
    def annotated_param(  # noqa: C901
        cls,
        annotation: Any,
        code: Optional[str] = None,
        matcher: Optional[Callable[[Any], bool]] = None,
        child_can_reuse_code: bool = False,
    ):
        """The decorator to register a type annotation for this function
        wrapper

        :param annotation: the type annotation
        :param code: the single char code to represent this type annotation
            , defaults to None, meaning it will try to use its parent class'
            code, this is allowed only if ``child_can_reuse_code`` is set to
            True on the parent class.
        :param matcher: a function taking in a type annotation and decide
            whether it is acceptable by the :class:`~.AnnotatedParam`
            , defaults to None, meaning it will just do a simple ``==`` check.
        :param child_can_reuse_code: whether the derived types of the current
            AnnotatedParam can reuse the code (if not specifying a new code)
            , defaults to False
        """

        def _func(tp: Type["AnnotatedParam"]) -> Type["AnnotatedParam"]:
            if not issubclass(tp, AnnotatedParam):
                raise InvalidOperationError(f"{tp} is not a subclass of AnnotatedParam")

            if matcher is not None:
                _matcher = matcher
            else:
                anno = annotation

                def _m(a: Any) -> bool:
                    return compare_annotations(a, anno, compare_origin=True)

                _matcher = _m

            tp._annotation = annotation
            if code is not None:
                tp._code = code
            else:
                tp._code = tp.__bases__[0]._code
            if tp._code in cls._REGISTERED_CODES:
                _allow_tp = cls._REGISTERED_CODES[tp._code]
                if (
                    _allow_tp is not None  # implies parent allows reusing the code
                    and inspect.isclass(tp)
                    and issubclass(tp, _allow_tp)
                ):
                    pass
                else:
                    for _ptp, _a, _c, _ in cls._REGISTERED:
                        if _c == tp._code:
                            if str(_ptp) != str(tp):
                                # This is to avoid a cyclic edge case
                                # If the first time import fugue fails, then because
                                # _REGISTERED_CODES is no longer empty, the second call
                                # could re-register the same classes which will cause
                                # exceptions.
                                #
                                # This trick ensures if there were duplication on the
                                # first try then in the second try it still fails at
                                # the same place.
                                #
                                # If import succeeded, this code will never be hit.

                                raise InvalidOperationError(
                                    f"param code {_c} is already registered by {_ptp}"
                                    f" {_a} so can't be used by {tp} {annotation}"
                                )
            else:
                if child_can_reuse_code and inspect.isclass(tp):
                    cls._REGISTERED_CODES[tp._code] = tp
                else:
                    cls._REGISTERED_CODES[tp._code] = None
            cls._REGISTERED.append((tp, annotation, code, _matcher))

            return tp

        return _func

    @classmethod
    def parse_annotation(
        cls,
        annotation: Any,
        param: Optional[inspect.Parameter] = None,
        none_as_other: bool = True,
    ) -> "AnnotatedParam":
        if annotation == type(None):  # noqa: E721
            return OtherParam(param) if none_as_other else NoneParam(param)
        if annotation == inspect.Parameter.empty:
            if param is not None and param.kind == param.VAR_POSITIONAL:
                return PositionalParam(param)
            if param is not None and param.kind == param.VAR_KEYWORD:
                return KeywordParam(param)
            return OtherParam(param) if none_as_other else NoneParam(param)

        load_entry_point(cls._ENTRYPOINT)

        for tp, _, _, matcher in cls._REGISTERED:
            if matcher(annotation):
                return tp(param)

        if param is not None and param.kind == param.VAR_POSITIONAL:
            return PositionalParam(param)
        if param is not None and param.kind == param.VAR_KEYWORD:
            return KeywordParam(param)
        return OtherParam(param)


class AnnotatedParam:
    """An abstraction of annotated parameter"""

    def __init__(self, param: Optional[inspect.Parameter]):
        if param is not None:
            self.required = param.default == inspect.Parameter.empty
            self.default = param.default
        else:
            self.required, self.default = True, None
        self.annotation: Any = getattr(self.__class__, "_annotation")  # noqa
        self.code: str = getattr(self.__class__, "_code")  # noqa

    def __repr__(self) -> str:
        return str(self.annotation)


def function_wrapper(entrypoint: Optional[str]):
    """The decorator to register a new :class:`~.FunctionWrapper` type.

    :param entrypoint: the entrypoint to load in setup.py in order to
        find the registered :class:`~.AnnotatedParam` under this
        :class:`~.FunctionWrapper`
    """

    def _func(tp: Type[FunctionWrapper]) -> Type[FunctionWrapper]:
        if not issubclass(tp, FunctionWrapper):
            raise InvalidOperationError(f"{tp} is not a subclass of FunctionWrapper")

        setattr(tp, "_REGISTERED", list(FunctionWrapper._REGISTERED))  # noqa
        setattr(  # noqa
            tp, "_REGISTERED_CODES", dict(FunctionWrapper._REGISTERED_CODES)
        )
        setattr(tp, "_ENTRYPOINT", entrypoint)  # noqa

        return tp

    return _func


@FunctionWrapper.annotated_param("NoneType", "n", lambda a: False)
class NoneParam(AnnotatedParam):
    """The case where there is no annotation for a parameter"""

    pass


@FunctionWrapper.annotated_param("[Self]", "0", lambda a: False)
class SelfParam(AnnotatedParam):
    """For the self parameters in member functions"""

    pass


@FunctionWrapper.annotated_param("[Other]", "x", lambda a: False)
class OtherParam(AnnotatedParam):
    """Any annotation that is not recognized"""

    pass


@FunctionWrapper.annotated_param("[Positional]", "y", lambda a: False)
class PositionalParam(AnnotatedParam):
    """For positional parameters"""

    pass


@FunctionWrapper.annotated_param("[Keyword]", "z", lambda a: False)
class KeywordParam(AnnotatedParam):
    """For keyword parameters"""

    pass



# flake8: noqa
from triad.utils.assertion import assert_arg_not_none, assert_or_throw
from triad.utils.class_extension import extensible_class, extension_method
from triad.utils.dispatcher import (
    conditional_broadcaster,
    conditional_dispatcher,
    run_at_def,
)
from triad.utils.hash import to_uuid
from triad.utils.iter import make_empty_aware
from triad.utils.threading import SerializableRLock, run_once



import logging
import sys
from typing import Any, List

from .threading import run_once

from importlib.metadata import EntryPoint, entry_points

_IMPORTLIB_META_VERSION: Any = ()


@run_once(key_func=lambda name: name)
def load_entry_point(name: str) -> None:
    """Load dependencies and functions of a given entrypoint. For any
    given entrypoint name, it will be loaded only once in one process.

    :param name: the name of the entrypoint

    .. admonition:: Example

        Assume in ``setup.py``, you have:

        .. code-block:: python

            setup(
                ...,
                entry_points={
                    "my.plugins": [
                        "my = pkg2.module2"
                    ]
                },
            )

        And this is how you load ``my.plugins``:

        .. code-block:: python

            from triad.utils.entrypoints import load_entry_point

            load_entry_point("my.plugins")
    """
    for plugin in _entry_points_for(name):
        _load_plugin(name, plugin)


def _load_plugin(entrypoint: str, plugin: EntryPoint) -> None:
    logger = logging.getLogger(_load_plugin.__name__)
    try:
        res = plugin.load()
        if callable(res):
            res()
        logger.debug("loaded %s %s", entrypoint, plugin)
    except Exception as e:  # pragma: no cover
        logger.debug("failed to load %s %s: %s", entrypoint, plugin, e)


def _entry_points_for(key: str) -> List[EntryPoint]:  # pragma: no cover
    if sys.version_info >= (3, 10) or _IMPORTLIB_META_VERSION >= (3, 6):
        return entry_points().select(group=key)  # type: ignore
    else:
        return entry_points().get(key, {})  # type: ignore
